/*
 * kgraft_patch_bsc1017589
 *
 * Fix for CVE-2016-9806, bsc#1017589
 *
 *  Upstream commit:
 *  92964c79b357 ("netlink: Fix dump skb leak/double free")
 *
 *  SLE12 commit:
 *  75ee08474d8faa4be52dc59470ad3743933d5f1c
 *
 *  SLE12-SP1 commit:
 *  12eaef947b92d094652cf8e10ef05cf18f2d74d8
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>

#include "local.h"

static void (*kgr_netlink_skb_set_owner_r)(struct sk_buff *skb, struct sock *sk);
static int (*kgr__netlink_sendskb)(struct sock *sk, struct sk_buff *skb);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "netlink_skb_set_owner_r", (void *)&kgr_netlink_skb_set_owner_r },
	{ "__netlink_sendskb", (void *)&kgr__netlink_sendskb },
};

int kgr_netlink_dump(struct sock *sk)
{
	struct netlink_sock *nlk = kgr_nlk_sk(sk);
	struct netlink_callback *cb;
	struct sk_buff *skb = NULL;
	struct nlmsghdr *nlh;
	/*
	 * Fix CVE-2016-9806
	 *  +1 line
	 */
	struct module *module;
	int len, err = -ENOBUFS;
	int alloc_size;

	mutex_lock(nlk->cb_mutex);
	if (!nlk->cb_running) {
		err = -EINVAL;
		goto errout_skb;
	}

	cb = &nlk->cb;
	alloc_size = max_t(int, cb->min_dump_alloc, NLMSG_GOODSIZE);

	if (!kgr_netlink_rx_is_mmaped(sk) &&
	    atomic_read(&sk->sk_rmem_alloc) >= sk->sk_rcvbuf)
		goto errout_skb;
	skb = netlink_alloc_skb(sk, alloc_size, nlk->portid, GFP_KERNEL);
	if (!skb)
		goto errout_skb;
	kgr_netlink_skb_set_owner_r(skb, sk);

	len = cb->dump(skb, cb);

	if (len > 0) {
		mutex_unlock(nlk->cb_mutex);

		if (sk_filter(sk, skb))
			kfree_skb(skb);
		else
			kgr__netlink_sendskb(sk, skb);
		return 0;
	}

	nlh = nlmsg_put_answer(skb, cb, NLMSG_DONE, sizeof(len), NLM_F_MULTI);
	if (!nlh)
		goto errout_skb;

	nl_dump_check_consistent(cb, nlh);

	memcpy(nlmsg_data(nlh), &len, sizeof(len));

	if (sk_filter(sk, skb))
		kfree_skb(skb);
	else
		kgr__netlink_sendskb(sk, skb);

	if (cb->done)
		cb->done(cb);

	nlk->cb_running = false;
	/*
	 * Fix CVE-2016-9806
	 *  +2 lines
	 */
	module = cb->module;
	skb = cb->skb;
	mutex_unlock(nlk->cb_mutex);
	/*
	 * Fix CVE-2016-9806
	 *  -module_put(cb->module);
	 *  -consume_skb(cb->skb);
	 *  +2 lines
	 */
	module_put(module);
	consume_skb(skb);
	return 0;

errout_skb:
	mutex_unlock(nlk->cb_mutex);
	kfree_skb(skb);
	return err;
}

static int kgr_patch_bsc1017589_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1017589_init(void)
{
	return kgr_patch_bsc1017589_kallsyms();
}
